from __future__ import annotations
from io import BytesIO
from typing import Protocol, List

import xlsxwriter
from xlsxwriter.format import Format
from xlsxwriter.worksheet import Worksheet


class WorkBook(Protocol):
    def add_format(self, format) -> Format:
        ...

    def worksheets(self) -> List[Worksheet]:
        ...

    def add_worksheet(self, name=None) -> Worksheet:
        ...


class BytesIOWorkBook:
    def __init__(self):
        self.io = BytesIO()
        self.workbook = xlsxwriter.Workbook(self.io)

    def add_format(self, format):
        return self.workbook.add_format(format)

    def close(self):
        self.workbook.close()
        self.io.seek(0)

    def read(self):
        data = self.io.read()
        self.io.close()
        return data

    def worksheets(self):
        return self.workbook.worksheets()

    def add_worksheet(self, name=None):
        return self.workbook.add_worksheet(name)


class Writer:
    def __init__(
        self,
        workbook: WorkBook,
        worksheet: Worksheet,
        global_format=None,
    ):

        self.workbook = workbook
        self.worksheet = worksheet
        self.global_format = self.workbook.add_format(global_format) if global_format else None
        self.global_format_dict = global_format or {}

    def _calc_format(self, cell_format):
        if cell_format is None:
            cell_format = self.global_format
        elif isinstance(cell_format, dict):
            cell_format = self.workbook.add_format(
                {**self.global_format_dict, **cell_format}
            )
        return cell_format

    def _path_args(self, args):
        last_arg = args[-1]
        if isinstance(last_arg, dict):
            args = list(args[:-1]) + [self._calc_format(last_arg)]
        return args

    def write(self, *args):
        args = self._path_args(args)
        self.worksheet.write(*args)

    def merge_range(self, *args):
        args = self._path_args(args)
        self.worksheet.merge_range(*args)

    def insert_image(self, *args):
        self.worksheet.insert_image(*args)



from __future__ import annotations
from io import BytesIO

from xlsxwriter import Workbook
from xlsxwriter.worksheet import Worksheet

from typing import Union, List

from .nodes import Box, BoxInstance, Col
from .visitors.printer import print_visitor
from .visitors.writer import writer_visitor
from .writer import Writer, BytesIOWorkBook


class Sheet:
    def __init__(
        self,
        root: Union[Box, List[Box]],
        start_row: int = 0,
        start_col: int = 0,
        name=None,
    ):
        if isinstance(root, list):
            root = Col(children=root)
        BoxInstance(root, start_row, start_col, None)
        self.root = root
        self.name = name

    @classmethod
    def attach_to_exist_worksheet(
        cls,
        workbook: Workbook,
        worksheet: Worksheet,
        root: Box,
        start_row: int = 0,
        start_col: int = 0,
    ):
        sheet = cls(root, start_row, start_col)
        writer = Writer(workbook, worksheet)
        visitor = writer_visitor(writer)
        sheet.root.accept(visitor)
        return writer

    def write_to_bytes_io(self) -> BytesIO:
        workbook = BytesIOWorkBook()
        worksheet = workbook.add_worksheet()
        writer = Writer(workbook, worksheet)
        visitor = writer_visitor(writer)
        self.root.accept(visitor)
        workbook.close()
        return workbook.io

    # old compat
    def write_to_bytesio(self):
        return self.write_to_bytes_io()

    def write(self, filename: str):
        io = self.write_to_bytes_io()
        with open(filename, "wb") as f:
            f.write(io.read())

    def print(self):
        self.root.accept(print_visitor)

    def to_bytes_io(self):
        return self.write_to_bytes_io()
